
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include <gridmap.h>

gridmap *gm_init( int width, int height, double grid_size, size_t elm_size, void *zero )
{
	gridmap *gm;
	int x, y;

	gm = malloc( sizeof( gridmap ) );
	if( gm == NULL ) return NULL;

	gm->width = width;
	gm->height = height;
	gm->size = grid_size;
	gm->elm_size = elm_size;
	gm->gridmap = malloc( elm_size * width * height );
	if( gm->gridmap == NULL )
	{
		free( gm );
		return NULL;
	}

	gm->zero = malloc( elm_size );
	if( gm->zero == NULL )
	{
		free( gm->gridmap );
		free( gm );
		return NULL;
	}

	gm->cache = malloc( sizeof(struct _cache) * width * height );
	if( gm->cache == NULL )
	{
		free( gm->zero );
		free( gm->gridmap );
		free( gm );
		return NULL;
	}
	for( y = 0; y < height; y ++ )
	{
		for( x = 0; x < width; x ++ )
		{
			struct _cache *tmp;
			float dx, dy;
			tmp = gm->cache + ( gm->width * y + x );
			dy = (float)y - (float)gm->height/2.0;
			dx = (float)x - (float)gm->width/2.0;
			tmp->orient = atan2f( dy, dx );
			tmp->length = sqrtf( dx * dx + dy * dy ) * grid_size;
		}
	}
	
	gm->origin.sin = 0;
	gm->origin.cos = 1;
	gm->origin.ix = 0;
	gm->origin.iy = 0;
	gm->origin.x = 0;
	gm->origin.y = 0;
	gm->origin.theta = 0;

	return gm;
}

void gm_free( gridmap *gm )
{
	free( gm->cache );
	free( gm->zero );
	free( gm->gridmap );
	free( gm );
} 

void gm_clear( gridmap *gm )
{
	int i, j;
	for( i = 0; i < gm->height; i ++ ){
		for( j = 0; j < gm->width; j ++ ){
			memcpy( gm->gridmap + gm->elm_size * ( gm->width * i + j ), gm->zero, gm->elm_size );
		}
	}
} 

void gm_set_origin( gridmap *gm, double x, double y, double theta )
{
	int i, j;
	int i0, i1;
	int j0, j1;
	int ix, iy;
	int dropx, dropy;

	dropx = (int)( ( x - gm->origin.x ) / gm->size );
	if( dropx < -( gm->width / 2 ) ) dropx = -( gm->width / 2 );
	if( dropx > gm->width / 2 ) dropx = gm->width / 2;

	dropy = (int)( ( y - gm->origin.y ) / gm->size );
	if( dropy < -( gm->height / 2 ) ) dropy = -( gm->height / 2 );
	if( dropy > gm->height / 2 ) dropy = gm->height / 2;

	ix = (int)( x / gm->size );
	iy = (int)( y / gm->size );

	if( dropx < 0 ){
		j0 = ix - ( gm->width / 2 );
		j1 = ix - ( gm->width / 2 ) - dropx + 1;
	}else if( dropx > 0 ){
		j0 = ix + ( gm->width / 2 ) - dropx;
		j1 = ix + ( gm->width / 2 ) + 1;
	}else{
		j0 = ix + ( gm->width / 2 ) - 1;
		j1 = ix + ( gm->width / 2 ) + 1;
	}

	if( dropy < 0 ){
		i0 = iy - ( gm->height / 2 );
		i1 = iy - ( gm->height / 2 ) - dropy + 1;
	}else if( dropy > 0 ){
		i0 = iy + ( gm->height / 2 ) - dropy;
		i1 = iy + ( gm->height / 2 ) + 1;
	}else{
		i0 = iy + ( gm->height / 2 ) - 1;
		i1 = iy + ( gm->height / 2 ) + 1;
	}

	gm_addressing_normalize( gm, &j0, &i0 );
	gm_addressing_normalize( gm, &j1, &i1 );

	//printf("%d,%d %d,%d\n",i0,i1,j0,j1);

	for( i = i0; i != i1; i ++ ){
		for( j = 0; j < gm->width; j ++ ){
			memcpy( gm->gridmap + gm->elm_size * ( gm->width * i + j ), gm->zero, gm->elm_size );
		}
		if( i >= gm->height - 1 ) i = -1;
	}

	for( i = 0; i < gm->height; i ++ ){
		for( j = j0; j != j1; j ++ ){
			memcpy( gm->gridmap + gm->elm_size * ( gm->width * i + j ), gm->zero, gm->elm_size );
			if( j >= gm->width - 1 ) j = -1;
		}
	}

	gm_addressing_normalize( gm, &ix, &iy );
	gm->origin.sin = sinf( theta );
	gm->origin.cos = cosf( theta );
	gm->origin.ix = ix;
	gm->origin.iy = iy;
	gm->origin.x = x;
	gm->origin.y = y;
	gm->origin.theta = theta;
} 

void gm_addressing_normalize( gridmap *gm, int *ix, int *iy )
{
	*ix = ((int)*ix) % gm->width;
	if( *ix < 0 ) *ix += gm->width;
	*iy = ((int)*iy) % gm->height;
	if( *iy < 0 ) *iy += gm->height;
}

int gm_addressing( gridmap *gm, float x, float y, int *ix, int *iy )
{
	int lx, ly;
	int ret;
	
	ret = 0;
	
	lx = (int)( ( x * gm->origin.cos - y * gm->origin.sin ) / gm->size );
	ly = (int)( ( x * gm->origin.sin + y * gm->origin.cos ) / gm->size );

	// mapあふれ
	if( abs( lx ) >= gm->width / 2 ) ret = 1;
	if( abs( ly ) >= gm->height / 2 ) ret = 1;

	*ix = gm->origin.ix + lx;
	*iy = gm->origin.iy + ly;

	return ret;
}

void gm_get_location( gridmap *gm, float *x, float *y, int ix, int iy )
{
	float px, py;

	while( ix > gm->width / 2 ) ix -= gm->width;
	while( iy > gm->height / 2 ) iy -= gm->height;
	while( ix < -gm->width / 2 ) ix += gm->width;
	while( iy < -gm->height / 2 ) iy += gm->height;

	px = (float)ix * gm->size;
	py = (float)iy * gm->size;

	*x =   px * gm->origin.cos + py * gm->origin.sin;
	*y = - px * gm->origin.sin + py * gm->origin.cos;
}

void *gm_element( gridmap *gm, float x, float y )
{
	int ix, iy;

	if( gm_addressing( gm, x, y, &ix, &iy ) ){
		return NULL;
	}
	gm_addressing_normalize( gm, &ix, &iy );

	return gm->gridmap + gm->elm_size * ( gm->width * iy + ix );
}

void *gm_element_direct( gridmap *gm, int ix, int iy )
{
	ix = gm->origin.ix + ix;
	iy = gm->origin.iy + iy;
	gm_addressing_normalize( gm, &ix, &iy );

	return gm->gridmap + gm->elm_size * ( gm->width * iy + ix );
}

void gm_scan_polar( gridmap *gm, float *scan_data, int len, int (*cross) (void *) )
{
	int u, v, i;
	float laser_angle;

	laser_angle = M_PI * 2.0 / len;
	for( i = 0; i < len; i ++ )
	{
		scan_data[i] = FLT_MAX;
	}

	for( v = - gm->height / 2; v < gm->height / 2; v ++ )
	{
		for( u = - gm->width / 2; u < gm->width / 2; u ++ )
		{
			if( cross( gm_element_direct( gm, u, v ) ) )
			{
				// 衝突判定 真
				struct _cache *tmp;
				int laser_width;
				int laser_center;

				tmp = &( gm->cache[ gm->width * ( v + gm->height / 2 ) + ( u + gm->width / 2 ) ] );
				laser_width = abs( (int)( ( tmp->length * sinf( laser_angle ) ) / gm->size ) );
				laser_center = ( tmp->orient - gm->origin.theta ) / laser_angle - len / 2;

				while( laser_center < 0 ) laser_center += len;
				while( laser_center >= len ) laser_center -= len;

				for( i = laser_center - laser_width; i <= laser_center - laser_width; i ++ )
				{
					int i_normal;
					
					i_normal = i;
					while( i_normal < 0 ) i_normal += len;
					while( i_normal >= len ) i_normal -= len;
					if( scan_data[ i_normal ] > tmp->length )
						scan_data[ i_normal ] = tmp->length;
				}
			}
		}
	}
}

void gm_linear_scan( gridmap *gm, float x1, float y1, float x2, float y2, int (*scan) (void *, float, void *), void *userdata )
{
	int u, v;
	int w, h;
	int sx, sy;
	int ex, ey;
	int wh, hw;
	int inv;
	int ua, va;
	
	sx = (int)( ( x1 * gm->origin.cos - y1 * gm->origin.sin ) / gm->size );
	sy = (int)( ( x1 * gm->origin.sin + y1 * gm->origin.cos ) / gm->size );
	ex = (int)( ( x2 * gm->origin.cos - y2 * gm->origin.sin ) / gm->size );
	ey = (int)( ( x2 * gm->origin.sin + y2 * gm->origin.cos ) / gm->size );
	
	w = ex - sx;
	h = ey - sy;
	
	inv = 0;
	
	if( abs( w ) > abs( h ) )
	{
		if( w < 0 )
		{
			int tmp;
			
			tmp = sx; sx = ex; ex = tmp;
			tmp = sy; sy = ey; ey = tmp;
			w = ex - sx;
			h = ey - sy;
			inv = 1;
		}
		v = 0;
		wh = w/2;
		hw = h/2;
		for( u = 0; u <= w; u ++ )
		{
			void *cell;
			
			if( inv )
			{
				ua = w - u;
				va = abs(h) - abs(v);
			}
			else
			{
				ua = u;
				va = v;
			}
			
			cell = gm_element_direct( gm, sx + u, sy + v );
			/*if( cell )*/ scan( cell, sqrtf( (float)va * va + (float)ua * ua ) * gm->size, userdata );
			
			if( h > 0 )
			{
				hw += h;
				if( wh < hw )
				{
					wh += w;
					v ++;
				}
			}
			else
			{
				hw += h;
				if( wh < -hw )
				{
					wh += w;
					v --;
				}
			}
		}
	}
	else
	{
		if( h < 0 )
		{
			int tmp;
			
			tmp = sx; sx = ex; ex = tmp;
			tmp = sy; sy = ey; ey = tmp;
			w = ex - sx;
			h = ey - sy;
			inv = 1;
		}
		u = 0;
		hw = h/2;
		wh = w/2;
		for( v = 0; v <= h; v ++ )
		{
			void *cell;
			
			if( inv )
			{
				va = h - v;
				ua = abs(w) - abs(u);
			}
			else
			{
				va = v;
				ua = u;
			}
			
			cell = gm_element_direct( gm, sx + u, sy + v );
			/*if( cell )*/ scan( cell, sqrtf( (float)va * va + (float)ua * ua ) * gm->size, userdata );
			
			if( w > 0 )
			{
				wh += w;
				if( hw < wh )
				{
					hw += h;
					u ++;
				}
			}
			else
			{
				wh += w;
				if( hw < -wh )
				{
					hw += h;
					u --;
				}
			}
		}
	}
}

